#!/usr/bin/env python

import argparse
import logging
import os
import sys

from kubernetes import client
from kubernetes.config import list_kube_config_contexts, load_kube_config

from k8s_handle import config
from k8s_handle import settings
from k8s_handle import templating
from k8s_handle.exceptions import DeprecationError, ProvisioningError
from k8s_handle.filesystem import InvalidYamlError
from k8s_handle.k8s.deprecation_checker import ApiDeprecationChecker
from k8s_handle.k8s.provisioner import Provisioner

COMMAND_DEPLOY = 'deploy'
COMMAND_DESTROY = 'destroy'

log = logging.getLogger(__name__)
logging.basicConfig(level=settings.LOG_LEVEL, format=settings.LOG_FORMAT, datefmt=settings.LOG_DATE_FORMAT)


def handler_deploy(args):
    _handler_deploy_destroy(args, COMMAND_DEPLOY)


def handler_destroy(args):
    _handler_deploy_destroy(args, COMMAND_DESTROY)


def handler_apply(args):
    _handler_apply_delete(args, COMMAND_DEPLOY)


def handler_delete(args):
    _handler_apply_delete(args, COMMAND_DESTROY)


def handler_render(args):
    context = config.load_context_section(args.get('section'))
    templating.Renderer(
        settings.TEMPLATES_DIR,
        args.get('tags'),
        args.get('skip_tags')
    ).generate_by_context(context)


def _handler_deploy_destroy(args, command):
    context = config.load_context_section(args.get('section'))
    resources = templating.Renderer(
        settings.TEMPLATES_DIR,
        args.get('tags'),
        args.get('skip_tags')
    ).generate_by_context(context)

    if args.get('dry_run'):
        return

    _handler_provision(
        command,
        resources,
        config.PriorityEvaluator(args, context, os.environ),
        args.get('use_kubeconfig'),
        args.get('sync_mode'),
        args.get('show_logs')
    )


def _handler_apply_delete(args, command):
    _handler_provision(
        command,
        [os.path.join(settings.TEMP_DIR, args.get('resource'))],
        config.PriorityEvaluator(args, {}, os.environ),
        args.get('use_kubeconfig'),
        args.get('sync_mode'),
        args.get('show_logs')
    )


def _handler_provision(command, resources, priority_evaluator, use_kubeconfig, sync_mode, show_logs):
    kubeconfig_namespace = None

    if priority_evaluator.environment_deprecated():
        log.warning("K8S_HOST and K8S_CA environment variables support is deprecated "
                    "and will be discontinued in the future. Use K8S_MASTER_URI and K8S_CA_BASE64 instead.")

    # INFO rvadim: https://github.com/kubernetes-client/python/issues/430#issuecomment-359483997
    if use_kubeconfig:
        try:
            load_kube_config()
            kubeconfig_namespace = list_kube_config_contexts()[1].get('context').get('namespace')
        except Exception as e:
            raise RuntimeError(e)
    else:
        client.Configuration.set_default(priority_evaluator.k8s_client_configuration())

    settings.K8S_NAMESPACE = priority_evaluator.k8s_namespace_default(kubeconfig_namespace)
    log.info('Default namespace "{}"'.format(settings.K8S_NAMESPACE))

    if not settings.K8S_NAMESPACE:
        log.info("Default namespace is not set. "
                 "This may lead to provisioning error, if namespace is not set for each resource.")

    try:
        deprecation_checker = ApiDeprecationChecker(client.VersionApi().get_code().git_version[1:])

        for resource in resources:
            deprecation_checker.run(resource)
    except client.api_client.ApiException:
        log.warning("Error while getting API version, deprecation check will be skipped.")

    provisioner = Provisioner(command, sync_mode, show_logs)

    for resource in resources:
        provisioner.run(resource)


parser = argparse.ArgumentParser(description='CLI utility generate k8s resources by templates and apply it to cluster')
subparsers = parser.add_subparsers(dest="command")
subparsers.required = True

parser_target_config = argparse.ArgumentParser(add_help=False)
parser_target_config.add_argument('-s', '--section', required=True, type=str, help='Section to deploy from config file')
parser_target_config.add_argument('-c', '--config', required=False, help='Config file, default: config.yaml')
parser_target_config.add_argument('--tags', action='append', required=False,
                                  help='Only use templates tagged with these values')
parser_target_config.add_argument('--skip-tags', action='append', required=False,
                                  help='Only use templates whose tags do not match these values')

parser_target_resource = argparse.ArgumentParser(add_help=False)
parser_target_resource.add_argument('-r', '--resource', required=True, type=str,
                                    help='Resource spec path, absolute (started with slash) or relative from TEMP_DIR')

parser_deprecated = argparse.ArgumentParser(add_help=False)
parser_deprecated.add_argument('--dry-run', required=False, action='store_true',
                               help='Don\'t run kubectl commands. Deprecated, use "k8s-handle template" instead')

parser_provisioning = argparse.ArgumentParser(add_help=False)
parser_provisioning.add_argument('--sync-mode', action='store_true', required=False, default=False,
                                 help='Turn on sync mode and wait deployment ending')
parser_provisioning.add_argument('--tries', type=int, required=False, default=360,
                                 help='Count of tries to check deployment status')
parser_provisioning.add_argument('--retry-delay', type=int, required=False, default=5,
                                 help='Sleep between tries in seconds')
parser_provisioning.add_argument('--strict', action='store_true', required=False,
                                 help='Check existence of all env variables in config.yaml and stop if var is not set')
parser_provisioning.add_argument('--use-kubeconfig', action='store_true', required=False,
                                 help='Try to use kube config')
parser_provisioning.add_argument('--k8s-handle-debug', action='store_true', required=False,
                                 help='Show K8S client debug messages')

parser_logs = argparse.ArgumentParser(add_help=False)
parser_logs.add_argument('--show-logs', action='store_true', required=False, default=False, help='Show logs for jobs')
parser_logs.add_argument('--tail-lines', type=int, required=False, help='Lines of recent log file to display')

arguments_connection = parser_provisioning.add_argument_group()
arguments_connection.add_argument('--k8s-master-uri', required=False, help='K8S master to connect to')
arguments_connection.add_argument('--k8s-ca-base64', required=False, help='base64-encoded K8S certificate authority')
arguments_connection.add_argument('--k8s-token', required=False, help='K8S token to use')

parser_deploy = subparsers.add_parser(
    'deploy',
    parents=[parser_provisioning, parser_target_config, parser_logs, parser_deprecated],
    help='Do attempt to create specs from templates and deploy K8S resources of the selected section')
parser_deploy.set_defaults(func=handler_deploy)

parser_apply = subparsers.add_parser('apply', parents=[parser_provisioning, parser_target_resource, parser_logs],
                                     help='Do attempt to deploy K8S resource from the existing spec')
parser_apply.set_defaults(func=handler_apply)

parser_destroy = subparsers.add_parser('destroy',
                                       parents=[parser_provisioning, parser_target_config, parser_deprecated],
                                       help='Do attempt to destroy K8S resources of the selected section')
parser_destroy.set_defaults(func=handler_destroy)

parser_delete = subparsers.add_parser('delete', parents=[parser_provisioning, parser_target_resource],
                                      help='Do attempt to destroy K8S resource from the existing spec')
parser_delete.set_defaults(func=handler_delete)

parser_template = subparsers.add_parser('render', parents=[parser_target_config],
                                        help='Make resources from the template and config. '
                                             'Created resources will be placed into the TEMP_DIR')
parser_template.set_defaults(func=handler_render)


def main():
    # INFO furiousassault: backward compatibility rough attempt
    # must be removed later according to https://github.com/2gis/k8s-handle/issues/40
    deprecation_warnings = 0
    filtered_arguments = []

    for argument in sys.argv[1:]:
        if argument in ['--sync-mode=true', '--sync-mode=True', '--dry-run=true', '--dry-run=True']:
            deprecation_warnings += 1
            filtered_arguments.append(argument.split('=')[0])
            continue

        if argument in ['--sync-mode=false', '--sync-mode=False', '--dry-run=false', '--dry-run=False']:
            deprecation_warnings += 1
            continue

        filtered_arguments.append(argument)

    args, unrecognized_args = parser.parse_known_args(filtered_arguments)

    if deprecation_warnings or unrecognized_args:
        log.warning("Explicit true/false arguments to --sync-mode and --dry-run keys are deprecated "
                    "and will be discontinued in the future. Use these keys without arguments instead.")

    args_dict = vars(args)
    settings.CHECK_STATUS_TRIES = args_dict.get('tries')
    settings.CHECK_DAEMONSET_STATUS_TRIES = args_dict.get('tries')
    settings.CHECK_STATUS_TIMEOUT = args_dict.get('retry_delay')
    settings.CHECK_DAEMONSET_STATUS_TIMEOUT = args_dict.get('retry_delay')
    settings.GET_ENVIRON_STRICT = args_dict.get('strict')
    settings.COUNT_LOG_LINES = args_dict.get('tail_lines')
    settings.CONFIG_FILE = args_dict.get('config') or settings.CONFIG_FILE

    try:
        args.func(args_dict)
    except templating.TemplateRenderingError as e:
        log.error('Template generation error: {}'.format(e))
        sys.exit(1)
    except InvalidYamlError as e:
        log.error('{}'.format(e))
        sys.exit(1)
    except DeprecationError as e:
        log.error('Deprecation warning: {}'.format(e))
        sys.exit(1)
    except RuntimeError as e:
        log.error('RuntimeError: {}'.format(e))
        sys.exit(1)
    except ProvisioningError:
        sys.exit(1)

    print(r'''
                         _(_)_                          wWWWw   _
             @@@@       (_)@(_)   vVVVv     _     @@@@  (___) _(_)_
            @@()@@ wWWWw  (_)\    (___)   _(_)_  @@()@@   Y  (_)@(_)
             @@@@  (___)     `|/    Y    (_)@(_)  @@@@   \|/   (_)
              /      Y       \|    \|/    /(_)    \|      |/      |
           \ |     \ |/       | / \ | /  \|/       |/    \|      \|/
            \|//    \|///    \|//  \|/// \|///    \|//    |//    \|//
       ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^''')


if __name__ == '__main__':
    main()



import base64
import glob
import itertools
import logging
import os
from hashlib import sha256

import yaml
from jinja2 import Environment, FileSystemLoader, StrictUndefined
from jinja2.exceptions import TemplateNotFound, TemplateSyntaxError, UndefinedError

from k8s_handle import settings
from k8s_handle.exceptions import TemplateRenderingError

log = logging.getLogger(__name__)


def get_template_contexts(file_path):
    try:
        with open(file_path) as f:
            try:
                contexts = yaml.safe_load_all(f.read())
            except Exception as e:
                raise RuntimeError('Unable to load yaml file: {}, {}'.format(file_path, e))

            for context in contexts:
                if context is None:
                    continue  # Skip empty YAML documents
                if 'kind' not in context or context['kind'] is None:
                    raise RuntimeError('Field "kind" not found (or empty) in file "{}"'.format(file_path))
                if 'metadata' not in context or context['metadata'] is None:
                    raise RuntimeError('Field "metadata" not found (or empty) in file "{}"'.format(file_path))
                if 'name' not in context['metadata'] or context['metadata']['name'] is None:
                    raise RuntimeError('Field "metadata->name" not found (or empty) in file "{}"'.format(file_path))
                if 'spec' in context:
                    # INFO: Set replicas = 1 by default for replaces cases in Deployment and StatefulSet
                    if 'replicas' not in context['spec'] or context['spec']['replicas'] is None:
                        if context['kind'] in ['Deployment', 'StatefulSet']:
                            context['spec']['replicas'] = 1
                yield context
    except FileNotFoundError as e:
        raise RuntimeError(e)


def b64decode(string):
    res = base64.decodebytes(string.encode())
    return res.decode()


def b64encode(string):
    res = base64.b64encode(string.encode())
    return res.decode()


def hash_sha256(string):
    res = sha256()
    res.update(string.encode('utf-8'))
    return res.hexdigest()


def get_env(templates_dir):
    # https://stackoverflow.com/questions/9767585/insert-static-files-literally-into-jinja-templates-without-parsing-them
    def include_file(path):
        path = os.path.join(templates_dir, '../', path)
        output = []
        for file_path in glob.glob(path):
            with open(file_path, 'r') as f:
                output.append(f.read())
        return '\n'.join(output)

    env = Environment(
        undefined=StrictUndefined,
        loader=FileSystemLoader([templates_dir]))

    env.filters['b64decode'] = b64decode
    env.filters['b64encode'] = b64encode
    env.filters['hash_sha256'] = hash_sha256
    env.globals['include_file'] = include_file

    log.debug('Available templates in path {}: {}'.format(templates_dir, env.list_templates()))
    return env


class Renderer:
    def __init__(self, templates_dir, tags=None, tags_skip=None):
        self._templates_dir = templates_dir
        self._tags = tags
        self._tags_skip = tags_skip
        self._env = get_env(self._templates_dir)

    def _iterate_entries(self, entries, tags=None):
        if tags is None:
            tags = set()

        for entry in entries:
            entry["tags"] = self._get_template_tags(entry).union(tags)

            if "group" not in entry.keys():
                if not self._evaluate_tags(entry.get("tags"), self._tags, self._tags_skip):
                    continue
                yield entry

            for nested_entry in self._iterate_entries(entry.get("group", []), entry.get("tags")):
                yield nested_entry

    def generate_by_context(self, context):
        if context is None:
            raise RuntimeError('Can\'t generate templates from None context')

        templates = context.get('templates', [])
        if len(templates) == 0:
            templates = context.get('kubectl', [])
            if len(templates) == 0:
                return

        output = []
        for template in self._iterate_entries(templates):
            try:
                path = self._generate_file(template, settings.TEMP_DIR, context)
                log.info('File "{}" successfully generated'.format(path))
                output.append(path)
            except TemplateNotFound as e:
                raise TemplateRenderingError(
                    "Processing {}: template {} hasn't been found".format(template['template'], e.name))
            except (UndefinedError, TemplateSyntaxError) as e:
                raise TemplateRenderingError('Unable to render {}, due to: {}'.format(template, e))
        return output

    def _generate_file(self, item, directory, context):
        try:
            log.info('Trying to generate file from template "{}" in "{}"'.format(item['template'], directory))
            template = self._env.get_template(item['template'])
        except TemplateNotFound as e:
            log.info('Templates path: {}, available templates: {}'.format(self._templates_dir,
                                                                          self._env.list_templates()))
            raise e
        except KeyError:
            raise RuntimeError('Templates section doesn\'t have any template items')

        new_name = item['template'].replace('.j2', '')
        path = os.path.join(directory, new_name)

        try:
            if not os.path.exists(os.path.dirname(path)):
                os.makedirs(os.path.dirname(path))

            with open(path, 'w+') as f:
                f.write(template.render(context))

        except TemplateRenderingError:
            raise
        except (FileNotFoundError, PermissionError) as e:
            raise RuntimeError(e)

        return path

    @staticmethod
    def _get_template_tags(template):
        if 'tags' not in template:
            return set()

        tags = template['tags']

        if isinstance(tags, list):
            return set([i for i, _ in itertools.groupby(tags)])

        if isinstance(tags, str):
            return set(tags.split(','))

        raise TypeError('Unable to parse tags of "{}" template: unexpected type {}'.format(template, type(tags)))

    @staticmethod
    def _evaluate_tags(tags, only_tags, skip_tags):
        if only_tags is None and skip_tags is None:
            return True

        return tags.isdisjoint(skip_tags or []) and not tags.isdisjoint(only_tags or tags)



import json
import re


def split_str_by_capital_letters(item):
    # upper the first letter
    item = item[0].upper() + item[1:]
    # transform 'Service' to 'service', 'CronJob' to 'cron_job', 'TargetPort' to 'target_port', etc.
    return '_'.join(re.findall(r'[A-Z][^A-Z]*', item)).lower()


def add_indent(json_str):
    try:
        return json.dumps(json.loads(json_str), indent=4)
    except:  # NOQA
        return json_str



import atexit
import logging
import os
import tempfile

import yaml

from k8s_handle.exceptions import InvalidYamlError

# furiousassault RE: it's not a good practice to log from utility function
# maybe we should pass os.remove failure silently, it doesn't seem so important
log = logging.getLogger(__name__)


def load_yaml(path):
    try:
        with open(path) as f:
            return yaml.safe_load(f.read())
    except Exception as e:
        raise InvalidYamlError("file '{}' doesn't contain valid yaml: {}".format(
            path, e))


def write_file_tmp(data):
    def remove_file(file_path):
        try:
            os.remove(file_path)
        except Exception as e:
            log.warning('Unable to remove "{}", due to "{}"'.format(file_path, e))

    f = tempfile.NamedTemporaryFile(delete=False)
    f.write(data)
    f.flush()
    atexit.register(remove_file, f.name)
    return f.name



import os
from setuptools import setup, find_packages

readme_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    'README.md')


def get_content(path):
    with open(path, 'r') as f:
        return f.read()


setup(name='k8s-handle',
      version=os.environ.get('RELEASE_TAG', '0.0.0'),
      long_description=get_content(readme_path),
      long_description_content_type='text/markdown',
      description='Provisioning tool for Kubernetes apps',
      url='http://github.com/2gis/k8s-handle',
      author='Vadim Reyder',
      author_email='vadim.reyder@gmail.com',
      license='Apache 2.0',
      packages=find_packages(exclude=("tests",)),
      data_files=['requirements.txt'],
      entry_points={
          "console_scripts": [
              "k8s-handle=k8s_handle:main",
          ]
      },
      install_requires=get_content('requirements.txt').split('\n'),
      zip_safe=False)



import logging
from time import sleep

from kubernetes.client.models.v1_label_selector import V1LabelSelector
from kubernetes.client.models.v1_label_selector_requirement import V1LabelSelectorRequirement
from kubernetes.client.models.v1_resource_requirements import V1ResourceRequirements

from k8s_handle import settings
from k8s_handle.templating import get_template_contexts
from k8s_handle.transforms import split_str_by_capital_letters
from .adapters import Adapter

log = logging.getLogger(__name__)


class Provisioner:
    def __init__(self, command, sync_mode, show_logs):
        self.command = command
        self.sync_mode = False if show_logs else sync_mode
        self.show_logs = show_logs

    @staticmethod
    def _replicas_count_are_equal(replicas):
        replicas = [0 if r is None else r for r in replicas]  # replace all None to 0
        return all(r == replicas[0] for r in replicas)

    @staticmethod
    def _ports_are_equal(old_port, new_port):
        for new_key in new_port.keys():
            old_key = split_str_by_capital_letters(new_key)
            if getattr(old_port, old_key, None) != new_port.get(new_key, None):
                return False
        return True

    @staticmethod
    def _get_missing_items_in_metadata_field(items, metadata, field):
        if field not in metadata:
            result = [item for item in items if 'kubernetes.io' not in item]
        else:
            result = [item for item in items if
                      item not in metadata[field] and 'kubernetes.io' not in item]

        return result

    @staticmethod
    def _port_obj_to_str(port):
        if hasattr(port, 'name') and hasattr(port, 'port'):
            return '{} ({})'.format(port.name, port.port)
        return '{}'.format(port.port)

    def _notify_about_missing_items_in_template(self, items, missing_type):
        skull = r"""
       ___
    .-'   `-.
   /  \   /  \\     Please pay attention to service {type}s!
  .   o\ /o   .    The next {type}(s) missing in template:
  |___  ^  ___|    "{list}"
      |___|        They won\'t be deleted after service apply.
      |||||
                    """

        if len(items) != 0:
            if missing_type in ['port', ]:
                items = [self._port_obj_to_str(item) for item in items]

            log_text = skull.format(type=missing_type, list=', '.join(items))
            if settings.GET_ENVIRON_STRICT:
                raise RuntimeError(log_text)
            log.warning(log_text)

    @staticmethod
    def _is_job_complete(status):
        if status.failed is not None:
            raise RuntimeError('Job running failed')

        if status.conditions is not None:
            for condition in status.conditions:
                if condition.type == 'Complete':
                    return True
        else:
            return False

    def _get_missing_annotations_and_labels(self, old_metadata, new_metadata):
        missing_annotations = []
        missing_labels = []

        if hasattr(old_metadata, 'annotations') and old_metadata.annotations is not None:
            missing_annotations = self._get_missing_items_in_metadata_field(
                old_metadata.annotations, new_metadata, 'annotations')
        if hasattr(old_metadata, 'labels') and old_metadata.labels is not None:
            missing_labels = self._get_missing_items_in_metadata_field(
                old_metadata.labels, new_metadata, 'labels')

        return missing_annotations, missing_labels

    def _get_apply_ports(self, old_spec, new_spec):
        ports = []

        if hasattr(old_spec, 'ports') and old_spec.ports is not None:
            if 'ports' not in new_spec:
                return []

            new_ports = new_spec['ports']
            for old_port in old_spec.ports:
                res = [item for item in new_ports if item['port'] == old_port.port]

                if len(res) == 0:
                    log.warning('Port {} will be deleted'.format(old_port.port))
                    ports.append({'$patch': 'delete', 'port': old_port.port})

                if len(res) == 1:
                    new_port = self._add_defaults_to_port(res[0])
                    if not self._ports_are_equal(old_port, new_port):
                        ports.append(new_port)

            for new_port in new_ports:
                res = [item for item in old_spec.ports if item.port == new_port['port']]

                if len(res) == 0:
                    ports.append(new_port)

        return ports

    @staticmethod
    def _add_defaults_to_port(port):
        if 'name' not in port:
            port['name'] = None
        if 'nodePort' not in port:
            port['node_port'] = None
        if 'protocol' not in port:
            port['protocol'] = 'TCP'
        if 'targetPort' not in port:
            port['target_port'] = port['port']

        return port

    def run(self, file_path):
        if self.command == 'deploy':
            self._deploy_all(file_path)
        if self.command == 'destroy':
            self._destroy_all(file_path)

    def _is_pvc_specs_equals(self, old_obj, new_dict):
        for new_key in new_dict.keys():
            old_key = split_str_by_capital_letters(new_key)

            # if template has some new attributes
            if not hasattr(old_obj, old_key):
                return False
            old_value = getattr(old_obj, old_key)

            if isinstance(old_value, list) and \
                    isinstance(old_value[0], V1LabelSelectorRequirement):
                if len(old_value) != len(new_dict[new_key]):
                    return False
                for i in range(0, len(old_value)):
                    if not self._is_pvc_specs_equals(old_value[i], new_dict[new_key][i]):
                        return False

            elif isinstance(old_value, (V1ResourceRequirements, V1LabelSelector)):
                if not self._is_pvc_specs_equals(old_value, new_dict[new_key]):
                    return False

            elif old_value != new_dict[new_key]:
                log.error('{} != {}'.format(old_value, new_dict[new_key]))
                return False

        return True

    def _deploy_all(self, file_path):
        for template_body in get_template_contexts(file_path):
            self._deploy(template_body, file_path)

    def _deploy(self, template_body, file_path):
        kube_client = Adapter.get_instance(template_body)

        if not kube_client:
            raise RuntimeError(
                'Unknown apiVersion "{}" in template "{}"'.format(
                    template_body['apiVersion'],
                    file_path
                )
            )

        log.info('Using namespace "{}"'.format(kube_client.namespace))
        resource = kube_client.get()

        if resource is None:
            log.info('{} "{}" does not exist, create it'.format(template_body['kind'], kube_client.name))
            kube_client.create()
        else:
            log.info('{} "{}" already exists, replace it'.format(template_body['kind'], kube_client.name))
            parameters = {}

            if template_body['kind'] == 'Service':
                missing_annotations, missing_labels = \
                    self._get_missing_annotations_and_labels(resource.metadata, template_body['metadata'])
                parameters['ports'] = self._get_apply_ports(resource.spec, template_body['spec'])

                self._notify_about_missing_items_in_template(missing_annotations, 'annotation')
                self._notify_about_missing_items_in_template(missing_labels, 'label')

                if parameters['ports']:
                    log.info('Next ports will be applied: {}'.format(parameters['ports']))

            if template_body['kind'] == 'PersistentVolumeClaim':
                if self._is_pvc_specs_equals(resource.spec, template_body['spec']):
                    log.info('PersistentVolumeClaim is not changed')
                    return

            if template_body['kind'] == 'PersistentVolume':
                if resource.status.phase in ['Bound', 'Released']:
                    log.warning('PersistentVolume has "{}" status, skip replacing'.format(resource.status.phase))
                    return

            if template_body['kind'] in ['CustomResourceDefinition', 'PodDisruptionBudget']:
                parameters['resourceVersion'] = resource.metadata.resource_version

            kube_client.replace(parameters)

        if self.sync_mode:
            if template_body['kind'] == 'Deployment':
                self._wait_deployment_complete(kube_client,
                                               tries=settings.CHECK_STATUS_TRIES,
                                               timeout=settings.CHECK_STATUS_TIMEOUT)

            if template_body['kind'] == 'StatefulSet':
                self._wait_statefulset_complete(kube_client,
                                                tries=settings.CHECK_STATUS_TRIES,
                                                timeout=settings.CHECK_STATUS_TIMEOUT)

            # INFO: vadim.reyder Since Kubernetes version 1.6 all DaemonSets by default have
            # `updateStrategy.type`=`RollingUpdate`, so we wait for deploy only if `updateStrategy.type` != 'OnDelete'.
            # WARNING: We consciously skip case with kubernetes version < 1.6, due to it's very old.
            if template_body['kind'] == 'DaemonSet' and \
                    template_body.get('spec').get('updateStrategy', {}).get('type') != 'OnDelete':
                self._wait_daemonset_complete(kube_client,
                                              tries=settings.CHECK_STATUS_TRIES,
                                              timeout=settings.CHECK_STATUS_TIMEOUT)

            if template_body['kind'] == 'Job':
                return self._wait_job_complete(kube_client,
                                               tries=settings.CHECK_STATUS_TRIES,
                                               timeout=settings.CHECK_STATUS_TIMEOUT)

        if template_body['kind'] == 'Job' and self.show_logs:
            pod_name, pod_containers = self._get_pod_name_and_containers_by_selector(
                kube_client,
                template_body['metadata']['name'],
                tries=settings.CHECK_STATUS_TRIES,
                timeout=settings.CHECK_STATUS_TIMEOUT)

            log.info("Got pod name and pod containers {} {}".format(pod_name, pod_containers))

            if not pod_name:
                log.warning('Pod not found for showing logs')
                return

            is_successful = self._wait_pod_running(
                kube_client,
                pod_name,
                tries=settings.CHECK_STATUS_TRIES,
                timeout=settings.CHECK_STATUS_TIMEOUT)

            for pod_container in pod_containers:
                log.info('\n{}'.format(kube_client.read_pod_logs(pod_name, pod_container)))

            if not is_successful:
                raise RuntimeError('Job running failed')

    def _destroy_all(self, file_path):
        for template_body in get_template_contexts(file_path):
            self._destroy(template_body, file_path)

    def _destroy(self, template_body, file_path):
        kube_client = Adapter.get_instance(template_body)

        if not kube_client:
            raise RuntimeError(
                'Unknown apiVersion "{}" in template "{}"'.format(
                    template_body['apiVersion'],
                    file_path
                )
            )

        log.info('Using namespace "{}"'.format(kube_client.namespace))
        log.info('Trying to delete {} "{}"'.format(template_body['kind'], kube_client.name))
        response = kube_client.delete()

        if response is None:
            log.info("{} {} is not found".format(template_body['kind'], kube_client.name))
            return

        # custom objects api response is a simple dictionary without message field
        if hasattr(response, 'message') and response.message is not None:
            raise RuntimeError('{} "{}" deletion failed: {}'.format(
                template_body['kind'], kube_client.name, response.message))

        if isinstance(response, dict) and not response.get('metadata', {}).get('deletionTimestamp'):
            raise RuntimeError('{} "{}" deletion failed: {}'.format(template_body['kind'], kube_client.name, response))

        if self.sync_mode:
            self._wait_destruction_complete(kube_client, template_body['kind'],
                                            tries=settings.CHECK_STATUS_TRIES,
                                            timeout=settings.CHECK_STATUS_TIMEOUT)

        log.info('{} "{}" has been deleted'.format(template_body['kind'], kube_client.name))

    @staticmethod
    def _get_pod_name_and_containers_by_selector(kube_client, selector, tries, timeout):
        for i in range(0, tries):
            pod = kube_client.get_pods_by_selector(selector)

            if len(pod.items) == 1:
                log.info('Found pod "{}"'.format(pod.items[0].metadata.name))
                containers = [container.name for container in pod.items[0].spec.containers]
                return pod.items[0].metadata.name, containers
            else:
                if len(pod.items) == 0:
                    log.warning('No pods found by job-name={}, next attempt in {} sec.'.format(selector, timeout))
                else:
                    names = [pod.metadata.name for pod in pod.items]
                    log.warning('More than one pod found by job-name={}: {}, '
                                'next attempt in {} sec.'.format(selector, names, timeout))
            sleep(timeout)

        log.error('Problems with getting pod by selector job-name={} for {} tries'.format(selector, tries))
        return '', []

    def _wait_deployment_complete(self, kube_client, tries, timeout):
        for i in range(0, tries):
            sleep(timeout)
            deployment = kube_client.get()
            status = deployment.status

            replicas = [deployment.spec.replicas, status.replicas, status.available_replicas,
                        status.ready_replicas, status.updated_replicas]

            log.info('desiredReplicas = {}, updatedReplicas = {}, availableReplicas = {}'.
                     format(replicas[0], replicas[4], replicas[2]))
            if self._replicas_count_are_equal(replicas) and status.unavailable_replicas is None:
                log.info('Deployment completed on {} attempt'.format(i + 1))
                return
            else:
                log.info('Deployment not completed on {} attempt, next attempt in {} sec.'.format(i + 1, timeout))

        raise RuntimeError('Deployment not completed for {} tries'.format(tries))

    def _wait_statefulset_complete(self, kube_client, tries, timeout):
        for i in range(0, tries):
            sleep(timeout)
            statefulset = kube_client.get()
            status = statefulset.status

            current_revision = status.current_revision
            update_revision = status.update_revision
            replicas = [statefulset.spec.replicas, status.current_replicas, status.ready_replicas]

            log.info('Current revision {}, should be {}'.format(current_revision, update_revision))
            if current_revision == update_revision:
                log.info('desiredReplicas = {}, updatedReplicas = {}, availableReplicas = {}'.
                         format(replicas[0], replicas[1], replicas[2]))
                if self._replicas_count_are_equal(replicas):
                    log.info('StatefulSet completed on {} attempt'.format(i))
                    return
            else:
                log.info('StatefulSet not completed on {} attempt, next attempt in {} sec.'.format(i, timeout))

        raise RuntimeError('StatefulSet not completed for {} tries'.format(tries))

    def _wait_daemonset_complete(self, kube_client, tries, timeout):
        for i in range(0, tries):
            sleep(timeout)
            status = kube_client.get().status

            replicas = [status.desired_number_scheduled, status.number_available,
                        status.number_ready, status.updated_number_scheduled]
            log.info('desiredNodes = {}, availableNodes = {}, readyNodes = {}, updatedNodes = {}'.
                     format(replicas[0], replicas[1], replicas[2], replicas[3]))
            if self._replicas_count_are_equal(replicas) and status.number_unavailable is None:
                log.info('DaemonSet completed on {} attempt'.format(i))
                return
            else:
                log.info('DaemonSet not completed on {} attempt, next attempt in {} sec.'.format(i, timeout))

        raise RuntimeError('DaemonSet not completed for {} tries'.format(tries))

    def _wait_job_complete(self, kube_client, tries, timeout):
        for i in range(0, tries):
            sleep(timeout)
            status = kube_client.get().status
            if self._is_job_complete(status):
                log.info('Job completed on {} attempt'.format(i))
                return
            else:
                log.info('Job not completed on {} attempt, next attempt in {} sec.'.format(i, timeout))

        raise RuntimeError('Job not completed for {} tries'.format(tries))

    @staticmethod
    def _wait_pod_running(kube_client, pod_name, tries, timeout):
        for i in range(0, tries):
            status = kube_client.read_pod_status(pod_name)

            log.info('Pod "{}" status: {}'.format(pod_name, status.status.phase))
            if status.status.phase == 'Succeeded':
                return True
            if status.status.phase in ['Failed', 'Unknown']:
                return False
            sleep(timeout)

        raise RuntimeError('Pod "{}" not completed for {} tries'.format(pod_name, tries))

    @staticmethod
    def _wait_destruction_complete(kube_client, kind, tries, timeout):
        for i in range(0, tries):
            sleep(timeout)
            if kube_client.get() is None:
                log.info('{} destruction completed on {} attempt'.format(kind, i + 1))
                return
            else:
                log.info('{} destruction not completed on {} attempt, '
                         'next attempt in {} sec.'.format(kind, i + 1, timeout))

        raise RuntimeError('{} destruction not completed for {} tries'.format(kind, tries))



import copy


def merge(dict_x, dict_y):
    result = copy.deepcopy(dict_x)
    for key, value in dict_y.items():
        if isinstance(value, dict) and isinstance(result.get(key), dict):
            result[key] = merge(result[key], value)
            continue

        result[key] = value

    return result
